# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================


class BaseSignal:
    def __init__(self):
        pass


class SignalError(BaseSignal):
    def __init__(self, exc: Exception = None):
        super().__init__()
        self.__exc = exc

    def get_exception(self):
        return self.__exc


class SignalKill(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalNoOp(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalOperationInit(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalOperationStart(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalOperationStop(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalResourcesCreation(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalResourcesDeletion(BaseSignal):

    def __init__(self):
        super().__init__()


class SignalServicesStart(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalServicesStop(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalShutdown(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalTerminate(BaseSignal):
    def __init__(self):
        super().__init__()